from celery import shared_task, chain
from django.utils import timezone

from nodeconductor.core.tasks import save_error_message, transition, retry_if_false, Task, ErrorStateTransitionTask

from .models import Instance, Volume
from .backend import AWSBackendError


@shared_task(name='nodeconductor_aws.destroy')
@transition(Instance, 'begin_deleting')
@save_error_message
def destroy(instance_uuid, transition_entity=None):
    instance = transition_entity
    try:
        backend = instance.get_backend()
        backend.destroy_instance(instance)
    except:
        set_erred(instance_uuid)
        raise
    else:
        instance.decrease_backend_quotas_usage()
        instance.delete()


@shared_task(name='nodeconductor_aws.start')
def start(instance_uuid):
    chain(
        begin_starting.s(instance_uuid),
        wait_for_instance_state.si(instance_uuid, 'RUNNING'),
    ).apply_async(
        link=set_online.si(instance_uuid),
        link_error=set_erred.si(instance_uuid))


@shared_task(name='nodeconductor_aws.stop')
def stop(instance_uuid):
    chain(
        begin_stopping.s(instance_uuid),
        wait_for_instance_state.si(instance_uuid, 'STOPPED'),
    ).apply_async(
        link=set_offline.si(instance_uuid),
        link_error=set_erred.si(instance_uuid))


@shared_task(name='nodeconductor_aws.restart')
def restart(instance_uuid):
    Instance.objects.get(uuid=instance_uuid)
    chain(
        begin_restarting.s(instance_uuid),
        wait_for_instance_state.si(instance_uuid, 'RUNNING'),
    ).apply_async(
        link=set_online.si(instance_uuid),
        link_error=set_erred.si(instance_uuid))


@shared_task(max_retries=300, default_retry_delay=3)
@retry_if_false
def wait_for_instance_state(instance_uuid, state=''):
    instance = Instance.objects.get(uuid=instance_uuid)
    try:
        backend = instance.get_backend()
        manager = backend.get_manager(instance)
        backend_instance = manager.get_node(instance.backend_id)
    except AWSBackendError:
        return False
    return backend_instance.state == backend.State.fromstring(state)


@shared_task
@transition(Instance, 'begin_starting')
@save_error_message
def begin_starting(instance_uuid, transition_entity=None):
    instance = transition_entity
    backend = instance.get_backend()
    backend.start_instance(instance)


@shared_task
@transition(Instance, 'begin_stopping')
@save_error_message
def begin_stopping(instance_uuid, transition_entity=None):
    instance = transition_entity
    backend = instance.get_backend()
    backend.stop_instance(instance)


@shared_task
@transition(Instance, 'begin_restarting')
@save_error_message
def begin_restarting(instance_uuid, transition_entity=None):
    instance = transition_entity
    backend = instance.get_backend()
    backend.reboot_instance(instance)


@shared_task
@transition(Instance, 'set_online')
def set_online(instance_uuid, transition_entity=None):
    instance = transition_entity
    instance.start_time = timezone.now()

    backend = instance.get_backend()
    manager = backend.get_manager(instance)
    backend_instance = manager.get_node(instance.backend_id)
    instance.public_ips = backend_instance.public_ips

    instance.save(update_fields=['start_time', 'public_ips'])


@shared_task
@transition(Instance, 'set_offline')
def set_offline(instance_uuid, transition_entity=None):
    instance = transition_entity
    instance.start_time = None
    instance.save(update_fields=['start_time'])


@shared_task
@transition(Instance, 'set_erred')
def set_erred(instance_uuid, transition_entity=None):
    pass


class RuntimeStateException(Exception):
    pass


class PollRuntimeStateTask(Task):
    max_retries = 300
    default_retry_delay = 5

    def get_backend(self, instance):
        return instance.get_backend()

    def execute(self, instance, backend_pull_method, success_state, erred_state):
        backend = self.get_backend(instance)
        getattr(backend, backend_pull_method)(instance)
        instance.refresh_from_db()
        if instance.runtime_state not in (success_state, erred_state):
            self.retry()
        elif instance.runtime_state == erred_state:
            raise RuntimeStateException(
                'Instance %s (PK: %s) runtime state become erred: %s' % (instance, instance.pk, erred_state))


class SetInstanceErredTask(ErrorStateTransitionTask):
    """ Mark instance as erred and delete resources that were not created. """

    def execute(self, instance):
        super(SetInstanceErredTask, self).execute(instance)

        # delete volume if it were not created on backend,
        # mark as erred if creation was started, but not ended,
        volume = instance.volume_set.first()
        if volume.state == Volume.States.CREATION_SCHEDULED:
            volume.delete()
        elif volume.state == Volume.States.OK:
            pass
        else:
            volume.set_erred()
            volume.save(update_fields=['state'])



import six

from rest_framework import decorators, exceptions, viewsets

from nodeconductor.core.views import StateExecutorViewSet
from nodeconductor.structure import views as structure_views

from . import filters, models, serializers, executors


class AmazonServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.AWSService.objects.all()
    serializer_class = serializers.ServiceSerializer
    import_serializer_class = serializers.InstanceImportSerializer

    def get_import_context(self):
        return {'resource_type': self.request.query_params.get('resource_type')}

    def get_serializer_class(self):
        from nodeconductor.structure import SupportedServices

        if self.request.method == 'POST':
            resource_type = self.request.data.get('type')
            if resource_type == SupportedServices.get_name_for_model(models.Instance):
                return serializers.InstanceImportSerializer
            elif resource_type == SupportedServices.get_name_for_model(models.Volume):
                return serializers.VolumeImportSerializer
        return super(AmazonServiceViewSet, self).get_serializer_class()


class AmazonServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.AWSServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer


class RegionViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Region.objects.all()
    serializer_class = serializers.RegionSerializer
    filter_class = filters.RegionFilter
    lookup_field = 'uuid'


class ImageViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Image.objects.all()
    serializer_class = serializers.ImageSerializer
    filter_class = filters.ImageFilter
    lookup_field = 'uuid'


class SizeViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.Size.objects.all()
    serializer_class = serializers.SizeSerializer
    filter_class = filters.SizeFilter
    lookup_field = 'uuid'


class InstanceViewSet(structure_views.VirtualMachineViewSet):
    queryset = models.Instance.objects.all()
    serializer_class = serializers.InstanceSerializer
    create_executor = executors.InstanceCreateExecutor
    resize_executor = executors.InstanceResizeExecutor

    serializers = {
        'resize': serializers.InstanceResizeSerializer
    }

    def get_serializer_class(self):
        serializer = self.serializers.get(self.action)
        return serializer or super(InstanceViewSet, self).get_serializer_class()

    def perform_provision(self, serializer):
        instance = serializer.save()
        volume = instance.volume_set.first()

        self.create_executor.execute(
            instance,
            image=serializer.validated_data.get('image'),
            size=serializer.validated_data.get('size'),
            ssh_key=serializer.validated_data.get('ssh_public_key'),
            volume=volume
        )

    @decorators.detail_route(methods=['post'])
    @structure_views.safe_operation(valid_state=models.Instance.States.OK)
    def resize(self, request, instance, uuid=None):
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        new_size = serializer.validated_data.get('size')
        self.resize_executor.execute(instance, size=new_size)


class VolumeViewSet(six.with_metaclass(structure_views.ResourceViewMetaclass,
                                       structure_views.ResourceViewMixin,
                                       StateExecutorViewSet)):
    queryset = models.Volume.objects.all()
    serializer_class = serializers.VolumeSerializer
    create_executor = executors.VolumeCreateExecutor
    delete_executor = executors.VolumeDeleteExecutor

    serializers = {
        'attach': serializers.VolumeAttachSerializer
    }

    @decorators.detail_route(methods=['post'])
    @structure_views.safe_operation(valid_state=models.Volume.States.OK)
    def detach(self, request, volume, uuid=None):
        if not volume.instance:
            raise exceptions.ValidationError('Volume is already detached.')

        executors.VolumeDetachExecutor.execute(volume)

    @decorators.detail_route(methods=['post'])
    @structure_views.safe_operation(valid_state=models.Volume.States.OK)
    def attach(self, request, volume, uuid=None):
        serializer = self.get_serializer(volume, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        executors.VolumeAttachExecutor.execute(volume)

    def get_serializer_class(self):
        serializer = self.serializers.get(self.action)
        return serializer or super(VolumeViewSet, self).get_serializer_class()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import jsonfield.fields


def copy_external_ips(apps, schema_editor):
    Instance = apps.get_model('nodeconductor_aws', 'Instance')
    for instance in Instance.objects.iterator():
        instance.public_ips = instance.external_ips
        instance.save()


class Migration(migrations.Migration):

    dependencies = [
        ('nodeconductor_aws', '0002_remove_awsservice_name'),
    ]

    operations = [
        migrations.AddField(
            model_name='instance',
            name='private_ips',
            field=jsonfield.fields.JSONField(default=[], help_text='List of private IP addresses', blank=True),
        ),
        migrations.AddField(
            model_name='instance',
            name='public_ips',
            field=jsonfield.fields.JSONField(default=[], help_text='List of public IP addresses', blank=True),
        ),
        migrations.RunPython(copy_external_ips),
        migrations.RemoveField(
            model_name='instance',
            name='external_ips',
        ),
        migrations.RemoveField(
            model_name='instance',
            name='internal_ips',
        ),
    ]
